# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Pydoc-markdown is an extensible framework for generating API documentation,
with a focus on Python source code and the Markdown output format.
"""


from nr.databind.core import Collect, Field, ObjectMapper, Struct, UnionType
from nr.databind.json import JsonModule
from nr.stream import concat
from pydoc_markdown.interfaces import Loader, Processor, Renderer
from pydoc_markdown.reflection import ModuleGraph
from pydoc_markdown.contrib.loaders.python import PythonLoader
from pydoc_markdown.contrib.processors.filter import FilterProcessor
from pydoc_markdown.contrib.processors.crossref import CrossrefProcessor
from pydoc_markdown.contrib.processors.smart import SmartProcessor
from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer
from typing import Union
import logging
import yaml

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '3.1.1'

mapper = ObjectMapper(JsonModule())
logger = logging.getLogger(__name__)


class PydocMarkdown(Struct):
  loaders = Field([Loader], default=lambda: [PythonLoader()])
  processors = Field([Processor], default=lambda: [
    FilterProcessor(), SmartProcessor(), CrossrefProcessor()])
  renderer = Field(Renderer, default=MarkdownRenderer)
  unknown_fields = Field([str], default=[], hidden=True)

  def __init__(self, *args, **kwargs):
    super(PydocMarkdown, self).__init__(*args, **kwargs)
    self.graph = ModuleGraph()
    self.resolver = None

  def load_config(self, data: Union[str, dict]):
    """
    Loads a YAML configuration from *data*.

    Args:
      data (str, dict): If a string is specified, it is treated as the path
        to a YAML file.
    """

    filename = None
    if isinstance(data, str):
      filename = data
      logger.info('Loading configuration file "%s".', filename)
      with open(data) as fp:
        data = yaml.safe_load(fp)

    collector = Collect()
    result = mapper.deserialize(data, type(self), filename=filename, decorations=[collector])
    vars(self).update(vars(result))

    self.unknown_fields = list(concat((str(n.locator.append(u)) for u in n.unknowns)
      for n in collector.nodes))

  def load_modules(self):
    for loader in self.loaders:
      loader.load(self.graph)

  def process(self):
    if self.resolver is None:
      self.resolver = self.renderer.get_resolver(self.graph)
    for processor in self.processors:
      processor.process(self.graph, self.resolver)

  def render(self):
    if self.resolver is None:
      self.resolver = self.renderer.get_resolver(self.graph)
    self.renderer.process(self.graph, self.resolver)
    self.renderer.render(self.graph)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Provides the #PydocmdProcessor class which converts the Pydoc-Markdown
highlighting syntax into Markdown.
"""

import re

from nr.databind.core import Struct
from nr.interface import implements, override
from pydoc_markdown.interfaces import Processor

# TODO @NiklasRosenstein Figure out a way to mark text linking to other
#     objects so that they can be properly handled by the renderer.

@implements(Processor)
class PydocmdProcessor(Struct):

  @override
  def process(self, graph, resolver):
    graph.visit(self.process_node)

  def process_node(self, node):
    if not getattr(node, 'docstring', None):
      return
    lines = []
    codeblock_opened = False
    current_section = None
    for line in node.docstring.split('\n'):
      if line.startswith("```"):
        codeblock_opened = (not codeblock_opened)
      if not codeblock_opened:
        line, current_section = self._preprocess_line(line, current_section)
      lines.append(line)
    node.docstring = '\n'.join(lines)

  def _preprocess_line(self, line, current_section):
    match = re.match(r'# (.*)$', line)
    if match:
      current_section = match.group(1).strip().lower()
      line = re.sub(r'# (.*)$', r'__\1__\n', line)

    if current_section in ('arguments', 'parameters'):
      style = r'- __\1__:\3'
    elif current_section in ('attributes', 'members', 'raises'):
      style = r'- `\1`:\3'
    elif current_section in ('returns',):
      style = r'`\1`:\3'
    else:
      style = None
    if style:
      #                  | ident  | types     | doc
      line = re.sub(r'\s*([^\\:]+)(\s*\(.+\))?:(.*)$', style, line)

    # Rewrite the argument type in the parentheses.
    if current_section in ('arguments', 'parameters'):
      def sub(m):
        return '__{}__ (`{}`):'.format(m.group(1), m.group(2))
      line = re.sub(r'__(\w+)\s*\((.*?)\)__:', sub, line)

    return line, current_section



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Implements a renderer that produces Markdown output.
"""

from nr.databind.core import Field, Struct
from nr.interface import implements
from pydoc_markdown.interfaces import Renderer, Resolver
from pydoc_markdown.reflection import Object, Class, Data, Function, Module, ModuleGraph
from typing import Optional
import io
import sys


@implements(Renderer)
class MarkdownRenderer(Struct):
  #: Can be used to explicitly specify a file object to render to.
  fp = Field(object, default=None, hidden=True)

  #: The name of the file to render to. If no file is specified, it will
  #: render to stdout.
  filename = Field(str, default=None)

  #: The encoding of the output file. This is ignored when rendering to
  #: stdout.
  encoding = Field(str, default='utf8')

  #: If enabled, inserts anchors before Markdown headers to ensure that
  #: links to the header work. This is enabled by default.
  insert_header_anchors = Field(bool, default=True)

  #: Generate HTML headers instead of Mearkdown headers. This is disabled
  #: by default.
  html_headers = Field(bool, default=False)

  #: Render names in headers as code (using backticks or `<code>` tags,
  #: depending on #html_headers). This is enabled by default.
  code_headers = Field(bool, default=False)

  #: Generate descriptive class titles by adding the word "Objects" after
  #: the class name. This is enabled by default.
  descriptive_class_title = Field(bool, default=True)

  #: Generate descriptivie module titles by adding the word "Module" before
  #: the module name. This is enabled by default.
  descriptive_module_title = Field(bool, default=False)

  #: Add the class name as a prefix to method names. This class name is
  #: also rendered as code if #code_headers is enabled. This is enabled
  #: by default.
  add_method_class_prefix = Field(bool, default=False)

  #: Add the class name as a prefix to member names. This is enabled by
  #: default.
  add_member_class_prefix = Field(bool, default=False)

  #: Add the full module name as a prefix to the title of the header.
  #: This is disabled by default.
  add_full_prefix = Field(bool, default=False)

  #: If #add_full_prefix is enabled, this will result in the prefix to
  #: be wrapped in a `<sub>` tag.
  sub_prefix = Field(bool, default=False)

  #: Render the definition of data members as a code block. This is disabled
  #: by default.
  data_code_block = Field(bool, default=False)

  #: Max length of expressions. If this limit is exceeded, the remaining
  #: characters will be replaced with three dots. This is set to 100 by
  #: default.
  data_expression_maxlength = Field(int, default=100)

  #: Render the class signature as a code block. This includes the "class"
  #: keyword, the class name and its bases. This is enabled by default.
  classdef_code_block = Field(bool, default=True)

  #: Render the constructor signature in the class definition code block
  #: if its `__init__()` member is not visible.
  classdef_render_init_signature_if_needed = Field(bool, default=True)

  #: Render classdef and function signature blocks in the Python help()
  #: style.
  signature_python_help_style = Field(bool, default=False)

  #: Render the function signature as a code block. This includes the "def"
  #: keyword, the function name and its arguments. This is enabled by
  #: default.
  signature_code_block = Field(bool, default=True)

  #: Render the function signature in the header. This is disabled by default.
  signature_in_header = Field(bool, default=False)

  #: Include the "def" keyword in the function signature. This is enabled
  #: by default.
  signature_with_def = Field(bool, default=False)

  #: Render the class name in the code block for function signature. Note
  #: that this results in invalid Python syntax to be rendered. This is
  #: disabled by default.
  signature_class_prefix = Field(bool, default=False)

  #: Add the string "python" after the backticks for code blocks. This is
  #: enabled by default.
  code_lang = Field(bool, default=True)

  #: Render a table of contents at the beginning of the file.
  render_toc = Field(bool, default=False)

  #: The title of the "Table of Contents" header.
  render_toc_title = Field(str, default='Table of Contents')

  #: The maximum depth of the table of contents. Defaults to 2.
  toc_maxdepth = Field(int, default=2)

  #: Render module headers. This is enabled by default.
  render_module_header = Field(bool, default=True)

  #: Render docstrings as blockquotes. This is disabled by default.
  docstrings_as_blockquote = Field(bool, default=False)

  #: Use a fixed header level for every kind of API object. The individual
  #: levels can be defined with #header_level_by_type.
  use_fixed_header_levels = Field(bool, default=True)

  #: Fixed header levels by API object type.
  header_level_by_type = Field({int}, default={
    'Module': 1,
    'Class': 2,
    'Method': 4,
    'Function': 4,
    'Data': 4,
  })

  def _render_toc(self, fp, level, obj):
    if level > self.toc_maxdepth:
      return
    if obj.visible:
      object_id = self._generate_object_id(obj)
      fp.write('  ' * level + '* [{}](#{})\n'.format(self._escape(obj.name), object_id))
      level += 1
    for child in obj.members.values():
      self._render_toc(fp, level, child)

  def _render_header(self, fp, level, obj):
    object_id = self._generate_object_id(obj)
    if self.use_fixed_header_levels:
      level = self.header_level_by_type.get(type(obj).__name__, level)
    if self.insert_header_anchors and not self.html_headers:
      fp.write('<a name="{}"></a>\n'.format(object_id))
    if self.html_headers:
      header_template = '<h{0} id="{1}">{{title}}</h{0}>'.format(level, object_id)
    else:
      header_template = level * '#' + ' {title}'
    fp.write(header_template.format(title=self._get_title(obj)))
    fp.write('\n\n')

  def _format_function_signature(self, func: Function, override_name: str = None, add_method_bar: bool = True) -> str:
    parts = []
    for dec in func.decorators:
      parts.append('@{}{}\n'.format(dec.name, dec.args or ''))
    if self.signature_python_help_style and not func.is_method():
      parts.append('{} = '.format(func.path()))
    if func.is_async:
      parts.append('async ')
    if self.signature_with_def:
      parts.append('def ')
    if self.signature_class_prefix and (
        func.is_function() and func.parent and func.parent.is_class()):
      parts.append(func.parent.name + '.')
    parts.append((override_name or func.name))
    parts.append('(' + func.signature_args + ')')
    if func.return_:
      parts.append(' -> {}'.format(func.return_))
    result = ''.join(parts)
    if add_method_bar and func.is_method():
      result = '\n'.join(' | ' + l for l in result.split('\n'))
    return result

  def _format_classdef_signature(self, cls: Class) -> str:
    bases = ', '.join(map(str, cls.bases))
    if cls.metaclass:
      bases += ', metaclass=' + str(cls.metaclass)
    code = 'class {}({})'.format(cls.name, bases)
    if self.signature_python_help_style:
      code = cls.path() + ' = ' + code
    if self.classdef_render_init_signature_if_needed and (
        '__init__' in cls.members and not cls.members['__init__'].visible):
      code += ':\n |  ' + self._format_function_signature(
        cls.members['__init__'], override_name=cls.name, add_method_bar=False)
    return code

  def _format_data_signature(self, data: Data) -> str:
    expr = str(data.expr)
    if len(expr) > self.data_expression_maxlength:
      expr = expr[:self.data_expression_maxlength] + ' ...'
    return data.name + ' = ' + expr

  def _render_signature_block(self, fp, obj):
    if self.classdef_code_block and obj.is_class():
      code = self._format_classdef_signature(obj)
    elif self.signature_code_block and obj.is_function():
      code = self._format_function_signature(obj)
    elif self.data_code_block and obj.is_data():
      code = self._format_data_signature(obj)
    else:
      return
    fp.write('```{}\n'.format('python' if self.code_lang else ''))
    fp.write(code)
    fp.write('\n```\n\n')

  def _render_object(self, fp, level, obj):
    if not isinstance(obj, Module) or self.render_module_header:
      self._render_header(fp, level, obj)
    self._render_signature_block(fp, obj)
    if obj.docstring:
      lines = obj.docstring.split('\n')
      if self.docstrings_as_blockquote:
        lines = ['> ' + x for x in lines]
      fp.write('\n'.join(lines))
      fp.write('\n\n')

  def _render_recursive(self, fp, level, obj):
    if obj.visible:
      self._render_object(fp, level, obj)
      level += 1
    for member in obj.members.values():
      self._render_recursive(fp, level, member)

  def _render_graph(self, graph, fp):
    if self.render_toc:
      if self.render_toc_title:
        fp.write('# {}\n\n'.format(self.render_toc_title))
      for m in graph.modules:
        self._render_toc(fp, 0, m)
      fp.write('\n')
    for m in graph.modules:
      self._render_recursive(fp, 1, m)

  def _get_title(self, obj):
    title = obj.name
    if (self.add_method_class_prefix and obj.is_method()) or \
       (self.add_member_class_prefix and obj.is_data()):
      title = obj.parent.name + '.' + title
    elif self.add_full_prefix and not obj.is_method():
      title = obj.path()

    if obj.is_function():
      if self.signature_in_header:
        title += '(' + obj.signature_args + ')'

    if self.code_headers:
      if self.html_headers or self.sub_prefix:
        if self.sub_prefix and '.' in title:
          prefix, title = title.rpartition('.')[::2]
          title = '<sub>{}.</sub>{}'.format(prefix, title)
        title = '<code>{}</code>'.format(title)
      else:
        title = '`{}`'.format(title)
    else:
      title = self._escape(title)
    if isinstance(obj, Module) and self.descriptive_module_title:
      title = 'Module ' + title
    if isinstance(obj, Class) and self.descriptive_class_title:
      title += ' Objects'
    return title

  def _generate_object_id(self, obj):
    parts = []
    while obj:
      parts.append(obj.name)
      obj = obj.parent
    return '.'.join(reversed(parts))

  def _escape(self, s):
    return s.replace('_', '\\_').replace('*', '\\*')

  def render_to_string(self, graph: ModuleGraph):
    fp = io.StringIO()
    self._render_graph(graph, fp)
    return fp.getvalue()

  # Renderer

  def get_resolver(self, graph: ModuleGraph) -> Resolver:
    def resolver(scope: Object, ref: str) -> Optional[str]:
      target = graph.resolve_ref(scope, ref.split('.'))
      if target:
        return '#' + self._generate_object_id(target)
      return None
    return Resolver(resolver)

  def render(self, graph: ModuleGraph):
    if self.filename is None:
      self._render_graph(graph, self.fp or sys.stdout)
    else:
      with io.open(self.filename, 'w', encoding=self.encoding) as fp:
        self._render_graph(graph, fp)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Pydoc-Markdown is a renderer for Python API documentation in Markdown format.

With no arguments it will load the default configuration file. If the
*config* argument is specified, it must be the name of a configuration file
or a YAML formatted object for the configuration.
"""

from nr.databind.core import StructType
from pydoc_markdown import __version__, PydocMarkdown, static
from pydoc_markdown.contrib.loaders.python import PythonLoader
from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer
from pydoc_markdown.contrib.renderers.mkdocs import MkdocsRenderer
from pydoc_markdown.interfaces import Server
from pydoc_markdown.util.watchdog import watch_paths
from typing import List, Set, Union
import click
import logging
import os
import sys
import webbrowser
import yaml

config_filenames = [
  'pydoc-markdown.yml',
  'pydoc-markdown.yaml',
]
default_config_notice = (
  'Using this option will disable loading the default configuration file.')
logger = logging.getLogger(__name__)


class RenderSession:
  """
  Helper class to load the Pydoc-markdown configuration from a file, apply
  overrides and invoke the renderer.
  """

  def __init__(self,
      config: Union[None, dict, str],  #: Configuration object or file
      render_toc: bool = None,  #: Override the "render_toc" option in the MarkdownRenderer
      search_path: List[str] = None,  #: Override the search path in the Python loader
      modules: List[str] = None,  #: Override the modules in the Python loader
      packages: List[str] = None,  #: Override the packages in the Python loader
      py2: bool = None,  #: Override Python2 compatibility in the Python loader
      ) -> None:
    self.config = config
    self.render_toc = render_toc
    self.search_path = search_path
    self.modules = modules
    self.packages = packages
    self.py2 = py2

  def _apply_overrides(self, config: PydocMarkdown):
    """
    Applies overrides to the configuration.
    """

    # Update configuration per command-line options.
    if self.modules or self.packages or self.search_path or self.py2 is not None:
      loader = next(
        (l for l in config.loaders if isinstance(l, PythonLoader)), None)
      if not loader:
        error('no python loader found')
      if self.modules:
        loader.modules = self.modules
      if self.packages:
        loader.packages = self.packages
      if self.search_path:
        loader.search_path = self.search_path
      if self.py2 is not None:
        loader.print_function = not self.py2

    if self.render_toc is not None:
      # Find the #MarkdownRenderer field for this renderer.
      for field in config.renderer.__fields__.values():
        if isinstance(field.datatype, StructType) and \
            issubclass(field.datatype.struct_cls, MarkdownRenderer):
          markdown = getattr(config.renderer, field.name)
          break
      else:
        if isinstance(config.renderer, MarkdownRenderer):
          markdown = config.renderer
        else:
          error('renderer {!r} does not expose a MarkdownRenderer'
                .format(type(config.renderer).__name__))
      markdown.render_toc = self.render_toc

  def load(self) -> PydocMarkdown:
    """
    Loads the configuration and applies the overrides.
    """

    config = PydocMarkdown()
    if self.config:
      config.load_config(self.config)
    self._apply_overrides(config)

    if config.unknown_fields:
      logger.warning('Unknown configuration options: %s', ', '.join(config.unknown_fields))

    return config

  def render(self, config: PydocMarkdown) -> List[str]:
    """
    Kicks off the rendering process and returns a list of files to watch.
    """

    config.load_modules()
    config.process()
    config.render()

    watch_files = set(m.location.filename for m in config.graph.modules)
    if isinstance(self.config, str):
      watch_files.add(self.config)

    return list(watch_files)

  def run_server(self, config: PydocMarkdown, open_browser: bool = False):
    """
    Watches files for changes and (re-) starts a server process that
    serves an HTML page from the renderer output on the fly.
    """

    if not Server.provided_by(config.renderer):
      error('renderer {!r} cannot be used with --server'
            .format(type(config.renderer).__name__))

    observer, event, process = None, None, None
    watch_files = []

    try:
      while True:
        # Initial render or re-render if a file changed.
        if not event or event.is_set():
          if event:
            config = self.load()
          logger.info('Rendering.')
          watch_files = self.render(config)
          if observer:
            observer.stop()
          observer, event = watch_paths(watch_files)
          if process:
            process = config.renderer.reload_server(process)

        # If the process doesn't exist, start it.
        if process is None:
          logger.info('Starting MkDocs serve.')
          process = config.renderer.start_server()
          if open_browser:
            open_browser = False
            webbrowser.open(config.renderer.get_server_url())

        event.wait(0.5)
    finally:
      if observer:
        observer.stop()
      if process:
        process.terminate()


def error(*args):
  print('error:', *args, file=sys.stderr)
  sys.exit(1)


@click.command(help=__doc__)
@click.argument('config', required=False)
@click.version_option(__version__)
@click.option('--bootstrap', is_flag=True,
  help='Render the default configuration file into the current working directory and quit.')
@click.option('--bootstrap-mkdocs', is_flag=True,
  help='Render a template configuration file for generating MkDpcs files into the current '
       'working directory and quit.')
@click.option('--verbose', '-v', is_flag=True, help='Increase log verbosity.')
@click.option('--quiet', '-q', is_flag=True, help='Decrease the log verbosity.')
@click.option('--module', '-m', 'modules', metavar='MODULE', multiple=True,
  help='The module to parse and generated API documentation for. Can be '
       'specified multiple times. ' + default_config_notice)
@click.option('--package', '-p', 'packages', metavar='PACKAGE', multiple=True,
  help='The package to parse and generated API documentation for including '
       'all sub-packages and -modules. Can be specified multiple times. '
       + default_config_notice)
@click.option('--search-path', '-I', metavar='PATH', multiple=True,
  help='A directory to use in the search for Python modules. Can be '
       'specified multiple times. ' + default_config_notice)
@click.option('--py2/--py3', 'py2', default=None,
  help='Switch between parsing Python 2 and Python 3 code. The default '
       'is Python 3. Using --py2 will enable parsing code that uses the '
       '"print" statement. This is equivalent of setting the print_function '
       'option of the "python" loader to False. ' + default_config_notice)
@click.option('--render-toc/--no-render-toc', default=None,
  help='Enable/disable the rendering of the TOC in the "markdown" renderer.')
@click.option('--server', '-s', is_flag=True,
  help='Watch for file changes and re-render if needed and start the server '
       'for the configured renderer. This doesn\'t work for all renderers.')
@click.option('--open', '-o', 'open_browser', is_flag=True,
  help='Open the browser after starting the server with -s,--server.')
def cli(
    config,
    bootstrap,
    bootstrap_mkdocs,
    verbose,
    quiet,
    modules,
    packages,
    search_path,
    render_toc,
    py2,
    server,
    open_browser):

  if bootstrap and bootstrap_mkdocs:
    error('--bootstrap and --bootstrap-mkdocs are incompatible options')
  if bootstrap or bootstrap_mkdocs:
    if config or modules or packages or search_path or render_toc \
        or py2 or server or open_browser:
      error('--bootstrap must be used as a sole argument')
    existing_file = next((x for x in config_filenames if os.path.isfile(x)), None)
    if existing_file:
      error('file already exists: {!r}'.format(existing_file))
    filename = config_filenames[0]
    with open(filename, 'w') as fp:
      if bootstrap_mkdocs:
        fp.write(static.DEFAULT_MKDOCS_CONFIG)
      else:
        fp.write(static.DEFAULT_CONFIG)
    print('created', filename)
    return

  if open_browser and not server:
    error('--open can only be used with --server')

  load_implicit_config = not any((modules, packages, search_path, py2 is not None))

  # Initialize logging.
  if verbose is not None:
    if verbose:
      level = logging.INFO
    elif quiet:
      level = logging.ERROR
    else:
      level = logging.WARNING
    logging.basicConfig(format='[%(levelname)s - %(name)s]: %(message)s', level=level)

  # Load the configuration.
  if config and (config.lstrip().startswith('{') or '\n' in config):
    config = yaml.safe_load(config)
  if config is None and load_implicit_config:
    try:
      config = next((x for x in config_filenames if os.path.exists(x)))
    except StopIteration:
      error('config file not found.')

  session = RenderSession(
    config=config,
    render_toc=render_toc,
    search_path=search_path,
    modules=modules,
    packages=packages,
    py2=py2)

  pydocmd = session.load()
  if server:
    session.run_server(pydocmd, open_browser)
  else:
    session.render(pydocmd)


if __name__ == '__main__':
  cli()  # pylint: disable=no-value-for-parameter



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Provides a processor that implements various filter capabilities.
"""

from nr.databind.core import Field, Struct
from nr.interface import implements
from pydoc_markdown.interfaces import Processor
from pydoc_markdown.reflection import ModuleGraph


@implements(Processor)
class FilterProcessor(Struct):
  """
  The `filter` processor removes module and class members based on certain
  criteria.

  # Example

  ```py
  - type: filter
    expression: not name.startswith('_') and default()
  ```
  """

  expression = Field(str, default=None)
  documented_only = Field(bool, default=True)
  exclude_private = Field(bool, default=True)
  exclude_special = Field(bool, default=True)
  include_root_objects = Field(bool, default=True)

  SPECIAL_MEMBERS = ('__path__', '__annotations__', '__name__', '__all__')

  def process(self, graph, _resolver):
    graph.visit(self._process_member)

  def _process_member(self, node):
    def _check(node):
      if self.documented_only and not node.docstring:
        return False
      if self.exclude_private and node.name.startswith('_') and not node.name.endswith('_'):
        return False
      if self.exclude_special and node.name in self.SPECIAL_MEMBERS:
        return False
      return True

    if self.expression:
      scope = {'name': node.name, 'node': node, 'default': _check}
      if not eval(self.expression, scope):  # pylint: disable=eval-used
        node.visible = False

    if self.include_root_objects and (
        not node.parent or isinstance(node.parent, ModuleGraph)):
      return

    if not _check(node):
      node.visible = False



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
`pydoc_markdown.interfaces`
===========================

This module defines the interfaces that can to be implemented for
Pydoc-Markdown to implement custom loaders for documentation data,
processors or renderers.
"""

from nr.databind.core import  SerializeAs, UnionType
from nr.interface import Interface, default
from typing import Optional
from .reflection import ModuleGraph, Object
import subprocess


@SerializeAs(UnionType.with_entrypoint_resolver('pydoc_markdown.interfaces.Loader'))
class Loader(Interface):
  """
  This interface describes an object that is capable of loading documentation
  data. The location from which the documentation is loaded must be defined
  with the configuration class.
  """

  def load(self, graph):  # type: (ModuleGraph) -> None
    """
    Fill the #ModuleGraph.
    """


class LoaderError(Exception):
  pass


@SerializeAs(UnionType.with_entrypoint_resolver('pydoc_markdown.interfaces.Processor'))
class Processor(Interface):
  """
  A processor is an object that takes a #ModuleGraph object as an input and
  transforms it in an arbitrary way. This usually processes docstrings to
  convert from various documentation syntaxes to plain Markdown.
  """

  def process(self, graph, resolver):  # type: (ModuleGraph, Optional[Resolver]) -> None
    pass


class Resolver(Interface):
  """ A resolver can be used by a #Processor to replace cross references
  with a hyperlink. """

  def resolve_ref(self, scope: Object, ref: str) -> Optional[str]:
    pass


@SerializeAs(UnionType.with_entrypoint_resolver('pydoc_markdown.interfaces.Renderer'))
class Renderer(Processor):
  """
  A renderer is an object that takes a #ModuleGraph as an input and produces
  output files or writes to stdout. It may also expose additional command-line
  arguments. There can only be one renderer at the end of the processor chain.

  Note that sometimes a renderer may need to perform some processing before
  the render step. To keep the possibility open that a renderer may implement
  generic processing that could be used without the actual renderering
  functionality, #Renderer is a subclass of #Processor.
  """

  @default
  def process(self, graph, resolver):  # type: (ModuleGraph, Optional[Resolver]) -> None
    pass

  def render(self, graph):  # type: (ModuleGraph) -> None
    pass

  def get_resolver(self, graph):  # type: (ModuleGraph) -> Optional[Resolver]
    pass


class Server(Interface):
  """
  This interface describes an object that can start a server process for
  live-viewing generated documentation in the browser. #Renderer
  implementations may additionally implement this interface to advocate their
  compatibility with the `--server` and `--open` options of the pydoc-markdown
  CLI.
  """

  def get_server_url(self) -> str:
    ...

  def start_server(self) -> subprocess.Popen:
    ...

  @default
  def reload_server(self, process: subprocess.Popen) -> subprocess.Popen:
    """
    Called when the files generated by pydoc-markdown have been updated.
    This gives the implementation a chance to reload the server process.
    The default implementation returns the *process* unchanged. Returning
    #None will automatically call #start_server() afterwards.
    """

    return process



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Struct
from nr.interface import implements, override
from pydoc_markdown.contrib.processors.sphinx import generate_sections_markdown
from pydoc_markdown.interfaces import Processor
import re


@implements(Processor)
class GoogleProcessor(Struct):
  """
  This class implements the preprocessor for Google and PEP 257 docstrings.

  https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html
  https://www.python.org/dev/peps/pep-0257/
  """

  _param_res = [
    re.compile(r'^(?P<param>\S+):\s+(?P<desc>.+)$'),
    re.compile(r'^(?P<param>\S+)\s+\((?P<type>[^)]+)\):\s+(?P<desc>.+)$'),
    re.compile(r'^(?P<param>\S+)\s+--\s+(?P<desc>.+)$'),
    re.compile(
      r'^(?P<param>\S+)\s+\{\[(?P<type>\S+)\]\}\s+--\s+(?P<desc>.+)$'),
    re.compile(
      r'^(?P<param>\S+)\s+\{(?P<type>\S+)\}\s+--\s+(?P<desc>.+)$'),
  ]

  _keywords_map = {
    'Args:': 'Arguments',
    'Arguments:': 'Arguments',
    'Attributes:': 'Attributes',
    'Example:': 'Examples',
    'Examples:': 'Examples',
    'Keyword Args:': 'Arguments',
    'Keyword Arguments:': 'Arguments',
    'Methods:': 'Methods',
    'Note:': 'Notes',
    'Notes:': 'Notes',
    'Other Parameters:': 'Arguments',
    'Parameters:': 'Arguments',
    'Return:': 'Returns',
    'Returns:': 'Returns',
    'Raises:': 'Raises',
    'References:': 'References',
    'See Also:': 'See Also',
    'Todo:': 'Todo',
    'Warning:': 'Warnings',
    'Warnings:': 'Warnings',
    'Warns:': 'Warns',
    'Yield:': 'Yields',
    'Yields:': 'Yields',
  }

  def check_docstring_format(self, docstring: str) -> bool:
    for section_name in self._keywords_map:
      if section_name in docstring:
        return True
    return False

  @override
  def process(self, graph, _resolver):
    graph.visit(self.process_node)

  def process_node(self, node):
    if not node.docstring:
      return

    lines = []
    current_lines = []
    in_codeblock = False
    keyword = None

    def _commit():
      if keyword:
        generate_sections_markdown(lines, {keyword: current_lines})
      else:
        lines.extend(current_lines)
      current_lines.clear()

    for line in node.docstring.split('\n'):
      if line.startswith("```"):
        in_codeblock = not in_codeblock
        current_lines.append(line)
        continue

      if in_codeblock:
        current_lines.append(line)
        continue

      line = line.strip()
      if line in self._keywords_map:
        _commit()
        keyword = self._keywords_map[line]
        continue

      if keyword is None:
        lines.append(line)
        continue

      for param_re in self._param_res:
        param_match = param_re.match(line)
        if param_match:
          if 'type' in param_match.groupdict():
            current_lines.append(
              '- `{param}` _{type}_ - {desc}'.format(**param_match.groupdict()))
          else:
            current_lines.append(
              '- `{param}` - {desc}'.format(**param_match.groupdict()))
          break

      if not param_match:
        current_lines.append('  {line}'.format(line=line))

    _commit()
    node.docstring = '\n'.join(lines)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" Utils for Pydoc-Markdown using the Watchdog library. """

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from typing import List, Tuple
import logging
import os
import threading

logger = logging.getLogger(__name__)


class _CallbackEventHandler(FileSystemEventHandler):

  def __init__(self, callback, filter_paths=None):
    self._callback = callback
    self._filter_paths = filter_paths

  def on_any_event(self, event):
    if self._filter_paths and event.src_path not in self._filter_paths:
      return
    self._callback(event)


def watch_paths(
  paths: List[str],
  recursive: bool = False
) -> Tuple[Observer, threading.Event]:
  """ Creates an observer for the specified *paths* and returns it together
  with a #threading.Event object. The event will be set when event occurred.
  """

  paths = [os.path.abspath(os.path.normpath(x)) for x in paths]
  directories = set(os.path.dirname(x) for x in paths)

  event = threading.Event()
  event_handler = _CallbackEventHandler(lambda _: event.set(), paths)
  observer = Observer()

  for directory in directories:
    observer.schedule(event_handler, directory, recursive)
  observer.start()

  return observer, event



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
This module provides a plug&play config for renderers that produce multiple
output files and allow the user to define the pages with the API objects
to document.
"""

from nr.databind.core import Collection, Field, ProxyType, Struct
from pydoc_markdown.interfaces import Renderer
from pydoc_markdown.reflection import ModuleGraph
from typing import Iterable, Optional, List
import collections
import copy
import fnmatch
import logging
import os
import re
import shutil

logger = logging.getLogger(__name__)
Page = ProxyType()


class IterHierarchyItem(Struct):
  page = Field(Page)
  parent_chain = Field([Page])

  def filename(self,
      parent_dir: Optional[str],  #: Parent directory to join with
      suffix_with_dot: str,  #: Suffix of the generated filename
      index_name: str = 'index',
      skip_empty_pages: bool = True,
      ) -> Optional[str]:
    path = [p.name for p in self.parent_chain] + [self.page.name]
    if self.page.children:
      if skip_empty_pages and not self.page.contents and not self.page.source:
        return None
      path.append(index_name)
    filename = os.path.join(*path) + suffix_with_dot
    if parent_dir:
      filename = os.path.join(parent_dir, filename)
    return filename


@Page.implementation  # pylint: disable=function-redefined
class Page(Struct):
  """
  Metadata for a page that a renderer implementation should understand
  in order to produce multiple output files. The page hierarchy defines
  the site navigation as well as the generated files.
  """

  #: The name of the page. This is usually the filename without the suffix.
  #: It may have additional meaning depending on the renderer. If it is
  #: not specified, it will fall back to a slug version of the #title.
  name = Field(str, default=None)

  #: The title of the page.
  title = Field(str)

  #: If set, the page does not generate a file but instead is rendered as
  #: a link to the specified URL in the site navigation.
  href = Field(str, default=None)

  #: If set, the page is rendered using the contents of the specified
  #: file. Some renderers may modify the source or add a preamble.
  source = Field(str, default=None)

  #: A list of glob patterns that match the absolute unique names of API
  #: objects to include for rendering in the page.
  contents = Field([str], default=None)

  #: A list of pages that are children of this page.
  children = Field([Page], default=list)

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    if not self.name:
      self.name = re.sub(r'\s+', '-', self.title.lower())

  def iter_hierarchy(self, parent_chain: List['Page'] = None) -> Iterable[IterHierarchyItem]:
    if parent_chain is None:
      parent_chain = []
    yield IterHierarchyItem(self, parent_chain)
    for child in self.children:
      yield from child.iter_hierarchy(parent_chain + [self])

  def filtered_graph(self, graph: ModuleGraph) -> ModuleGraph:
    """
    Creates a copy of the module graph where only the API objects selected
    via #Page.contents are visible.
    """

    def _match(node):
      if self.contents:
        path = node.path()
        return any(fnmatch.fnmatch(path, x) for x in self.contents or ())
      return False

    def _update_nodes(x):
      x.visible = x.visible and _match(x)
      # Make sure all parents are visible as well.
      while x and x.visible:
        x = x.parent
        if x:
          x.visible = True

    clone = copy.deepcopy(graph)
    clone.visit(_update_nodes)
    return clone

  def render(
      self,
      filename: str,
      graph: ModuleGraph,
      renderer: Renderer
      ) -> None:
    """
    Renders the page by either copying the *source* to the specified
    *filename* or by rendering the *contents* from the *graph* using the
    specified *renderer*.

    Note that the *renderer* should be pre-configured to output to *filename*.
    """

    os.makedirs(os.path.dirname(filename), exist_ok=True)
    if self.source:
      logger.info('Writing "%s" (source: "%s")', filename, self.source)
      shutil.copyfile(self.source, filename)
    else:
      logger.info('Rendering "%s"', filename)
      renderer.render(self.filtered_graph(graph))


class Pages(Collection, list):
  item_type = Page

  def iter_hierarchy(self) -> Iterable[IterHierarchyItem]:
    for page in self:
      yield from page.iter_hierarchy()



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
This module provides the abstract representation of a code library. It is
generalised and intended to be usable for any language.
"""

from nr.databind.core import Field, Struct, ProxyType
from typing import List, Optional

Argument = ProxyType()
Decorator = ProxyType()
Expression = ProxyType()
Object = ProxyType()


class Location(Struct):
  filename = Field(str)
  lineno = Field(int)

  def __str__(self):
    return '{}:{}'.format(self.filename, self.lineno)


@Object.implementation  # pylint: disable=function-redefined
class Object(Struct):
  location = Field(Location, nullable=True)
  parent = Field(Object, default=None)
  name = Field(str)
  docstring = Field(str, default=None)
  members = Field(dict, default=dict)
  visible = Field(bool, default=True)

  def __init__(self, *args, **kwargs):
    super(Object, self).__init__(*args, **kwargs)
    if self.parent is not None:
      if self.name in self.parent.members:
        self.parent.members[self.name].remove()
      self.parent.members[self.name] = self

  def __repr__(self):
    v = ['{}={!r}'.format(k, getattr(self, k)) for k in self.__fields__
         if k not in ('location', 'parent', 'docstring', 'members')]
    return '{}({})'.format(type(self).__name__, ', '.join(v))

  def path(self, separator='.'):
    if not self.parent or isinstance(self.parent, ModuleGraph):
      return self.name
    else:
      return self.parent.path(separator) + separator + self.name

  def remove(self):
    if self.parent:
      assert self.parent.members[self.name] is self
      del self.parent.members[self.name]
      self.parent = None

  def is_module(self):
    return isinstance(self, Module)

  def is_class(self):
    return isinstance(self, Class)

  def is_data(self):
    return isinstance(self, Data)

  def is_function(self):
    return isinstance(self, Function)

  def is_method(self):
    if not self.parent:
      return False
    return isinstance(self, Function) and isinstance(self.parent, Class)

  def visit(self, func, allow_mutation=False):
    members = self.members.values()
    if allow_mutation:
      members = list(members)
    for child in members:
      child.visit(func, allow_mutation)
    func(self)


class Module(Object):
  pass


class Class(Object):
  decorators = Field([Decorator], nullable=True)
  bases = Field([Expression])
  metaclass = Field(Expression, nullable=True)


class Function(Object):
  is_async = Field(bool)
  decorators = Field([Decorator])
  args = Field([Argument])
  return_ = Field(Expression, nullable=True)

  @property
  def signature(self):
    return '{}({})'.format(self.name, Argument.format_arglist(self.args))

  @property
  def signature_args(self):
    if self.is_method and self.args and self.args[0].name == 'self':
      return Argument.format_arglist(self.args[1:])
    return Argument.format_arglist(self.args)


class Data(Object):
  expr = Field(Expression, default=None)
  annotation = Field(Expression, default=None)


@Decorator.implementation  # pylint: disable=function-redefined
class Decorator(Struct):
  name = Field(str)
  args = Field(Expression, default=None)


@Argument.implementation  # pylint: disable=function-redefined
class Argument(Struct):
  name = Field(str)
  annotation = Field(Expression, default=None)
  default = Field(Expression, default=None)
  type = Field(str)

  POS = 'pos'
  POS_REMAINDER = 'pos_remainder'
  KW_SEPARATOR = 'kw_separator'
  KW_ONLY = 'kw_only'
  KW_REMAINDER = 'kw_remainder'

  def __str__(self):
    if self.type == Argument.KW_SEPARATOR:
      return '*'
    parts = [self.name]
    if self.annotation:
      parts.append(': ' + str(self.annotation))
    if self.default:
      if self.annotation:
        parts.append(' ')
      parts.append('=')
    if self.default:
      if self.annotation:
        parts.append(' ')
      parts.append(str(self.default))
    if self.type == Argument.POS_REMAINDER:
      parts.insert(0, '*')
    elif self.type == Argument.KW_REMAINDER:
      parts.insert(0, '**')
    return ''.join(parts)

  @staticmethod
  def format_arglist(arglist):
    parts = []
    for arg in arglist:
      parts.append(str(arg))
    return ', '.join(parts)


@Expression.implementation  # pylint: disable=function-redefined
class Expression(Struct):
  text = Field(str)

  def __str__(self):
    return self.text


class ModuleGraph(Object):
  """
  Represents a collection of #Object#s, typically #Module#s.
  """

  # TODO (@NiklasRosenstein): Change "module" naming to "node".

  def __init__(self, nodes=()):
    super().__init__(
      location=None,
      parent=None,
      name='',
      docstring=None,
      members={})
    for node in nodes:
      self.add_module(node)

  @property
  def modules(self):
    return list(self.members.values())

  def add_module(self, node):
    self.members[node.name] = node
    node.parent = self

  def resolve_ref(self, scope: Object, ref: List[str]) -> Optional[Object]:
    """ Resolve a reference to another object from within the specified
    *scope*. Returns the target, or None if the target cannot be resolved
    from the ref.

    The *ref* must be a list of names that can be retrieved from the scope's
    members in order. If no scope is specified, it will be resolved inside
    the #ModuleGraph. """

    if not ref:
      return None

    def _single_resolve(scope, ref):
      for name in ref:
        if name not in scope.members:
          return None
        scope = scope.members[name]
      return scope

    while scope:
      target = _single_resolve(scope, ref)
      if target:
        return target
      scope = scope.parent

    fake = Object(location=None, name='fake', members={x.name: x for x in self.modules})
    return _single_resolve(fake, ref)


__all__ = [
  'Location',
  'Module',
  'Class',
  'Function',
  'Data',
  'Decorator',
  'Argument',
  'Expression'
]



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Provides the #SphinxProcessor that converts reST/Sphinx syntax to
markdown compatible syntax.
"""

from nr.databind.core import Struct
from nr.interface import implements, override
from pydoc_markdown.interfaces import Processor
import re


def generate_sections_markdown(lines, sections):
  for key, section in sections.items():
    if lines and lines[-1]:
      lines.append('')

    lines.extend(['**{}**:'.format(key), ''])  # add an extra line because of markdown syntax
    lines.extend(section)


@implements(Processor)
class SphinxProcessor(Struct):

  def check_docstring_format(self, docstring: str) -> bool:
    return ':param' in docstring or ':return' in docstring or \
      ':raise' in docstring

  @override
  def process(self, graph, _resolver):
    graph.visit(self.process_node)

  def process_node(self, node):
    if not node.docstring:
      return

    lines = []
    in_codeblock = False
    keyword = None
    components = {}

    for line in node.docstring.split('\n'):
      line = line.strip()

      if line.startswith("```"):
        in_codeblock = not in_codeblock

      line_codeblock = line.startswith('    ')

      if not in_codeblock and not line_codeblock:
        match = re.match(r'\s*:(?:arg|argument|param|parameter)\s+(\w+)\s*:(.*)?$', line)
        if match:
          keyword = 'Arguments'
          param = match.group(1)
          text = match.group(2)
          text = text.strip()

          component = components.setdefault(keyword, [])
          component.append('- `{}`: {}'.format(param, text))
          continue

        match = re.match(r'\s*:(?:return|returns)\s*:(.*)?$', line)
        if match:
          keyword = 'Returns'
          text = match.group(1)
          text = text.strip()

          component = components.setdefault(keyword, [])
          component.append(text)
          continue

        match = re.match('\\s*:(?:raises|raise)\\s+(\\w+)\\s*:(.*)?$', line)
        if match:
          keyword = 'Raises'
          exception = match.group(1)
          text = match.group(2)
          text = text.strip()

          component = components.setdefault(keyword, [])
          component.append('- `{}`: {}'.format(exception, text))
          continue

      if keyword is not None:
        components[keyword].append(line)
      else:
        lines.append(line)

    generate_sections_markdown(lines, components)
    node.docstring = '\n'.join(lines)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Loads Python source code into the Pydoc-markdown reflection format.
"""

import logging
import os
import pkgutil
import re
import textwrap
import sys

from lib2to3.refactor import RefactoringTool
from lib2to3.pgen2 import token
from lib2to3.pgen2.parse import ParseError
from lib2to3.pygram import python_symbols as syms
from lib2to3.pytree import Leaf, Node
from nr.databind.core import Field, Struct
from nr.interface import implements
from pydoc_markdown.interfaces import Loader, LoaderError
from pydoc_markdown.reflection import (
  Argument,
  Expression,
  Class,
  Data,
  Decorator,
  Function,
  Location,
  Module,
  ModuleGraph)

_REVERSE_SYMS = {v: k for k, v in vars(syms).items() if isinstance(v, int)}
_REVERSE_TOKEN = {v: k for k, v in vars(token).items() if isinstance(v, int)}

logger = logging.getLogger(__name__)


def dedent_docstring(s):
  lines = s.split('\n')
  lines[0] = lines[0].strip()
  lines[1:] = textwrap.dedent('\n'.join(lines[1:])).split('\n')
  return '\n'.join(lines).strip()


def find(predicate, iterable):
  for item in iterable:
    if predicate(item):
      return item
  return None


def parse_to_ast(code, filename, print_function=True):
  """
  Parses the string *code* to an AST with #lib2to3.
  """

  options = {'print_function': print_function}

  try:
    # NOTE (@NiklasRosenstein): Adding newline at the end as a ParseError
    #   could be raised without a trailing newline (tested in CPython 3.6
    #   and 3.7).
    return RefactoringTool([], options).refactor_string(code + '\n', filename)
  except ParseError as exc:
    raise ParseError(exc.msg, exc.type, exc.value, exc.context + (filename,))


def parse_file(code, filename, module_name=None, parser=None, **kwargs):
  """
  Creates a reflection of the Python source in the string *code*.
  """

  if code is None:
    with open(filename, 'r') as fp:
      code = fp.read()

  if parser is None:
    parser = Parser()

  return parser.parse(parse_to_ast(code, filename, **kwargs), filename, module_name)


class Parser:
  """
  A helper class that parses a Python AST to turn it into objects of the
  #pydoc_markdown.reflection module. Extracts docstrings from functions
  and single-line comments.
  """

  treat_singleline_comment_blocks_as_docstrings = True

  def parse(self, ast, filename, module_name=None):
    self.filename = filename  # pylint: disable=attribute-defined-outside-init

    if module_name is None:
      module_name = os.path.basename(filename)
      module_name = os.path.splitext(module_name)[0]

    docstring = self.get_docstring_from_first_node(ast, module_level=True)
    module = Module(self.location_from(ast), None, module_name, docstring)

    for node in ast.children:
      self.parse_declaration(module, node)

    return module

  def parse_declaration(self, parent, node, decorators=None):
    if node.type == syms.simple_stmt:
      assert not decorators
      stmt = node.children[0]
      if stmt.type in (syms.import_name, syms.import_from):
        # TODO @NiklasRosenstein handle import statements?
        pass
      elif stmt.type == syms.expr_stmt:
        return self.parse_statement(parent, stmt)
    elif node.type == syms.funcdef:
      return self.parse_funcdef(parent, node, False, decorators)
    elif node.type == syms.classdef:
      return self.parse_classdef(parent, node, decorators)
    elif node.type in (syms.async_stmt, syms.async_funcdef):
      child = node.children[1]
      if child.type == syms.funcdef:
        return self.parse_funcdef(parent, child, True, decorators)
    elif node.type == syms.decorated:
      assert len(node.children) == 2
      decorators = []
      if node.children[0].type == syms.decorator:
        decorator_nodes = [node.children[0]]
      elif node.children[0].type == syms.decorators:
        decorator_nodes = node.children[0].children
      else:
        assert False, node.children[0].type
      for child in decorator_nodes:
        assert child.type == syms.decorator, child.type
        decorators.append(self.parse_decorator(child))
      return self.parse_declaration(parent, node.children[1], decorators)
    return None

  def _split_statement(self, stmt):
    """
    Parses a statement node into three lists, consisting of the leaf nodes
    that are the name(s), annotation and value of the expression. The value
    list will be empty if this is not an assignment statement (but for example
    a plain expression).
    """

    def _parse(stack, current, stmt):
      for child in stmt.children:
        if not isinstance(child, Node) and child.value == '=':
          stack.append(current)
          current = ('value', [])
        elif not isinstance(child, Node) and child.value == ':':
          stack.append(current)
          current = ('annotation', [])
        elif isinstance(child, Node) and child.type == syms.annassign:
          _parse(stack, current, child)
        else:
          current[1].append(child)
      stack.append(current)
      return stack

    result = dict(_parse([], ('names', []), stmt))
    return result.get('names', []), result.get('annotation', []), result.get('value', [])

  def parse_statement(self, parent, stmt):
    names, annotation, value = self._split_statement(stmt)
    if value or annotation:
      docstring = self.get_statement_docstring(stmt)
      expr = Expression(self.nodes_to_string(value)) if value else None
      annotation = Expression(self.nodes_to_string(annotation)) if annotation else None
      assert names
      for name in names:
        name = self.nodes_to_string([name])
        data = Data(
          self.location_from(stmt),
          parent,
          name,
          docstring,
          expr=expr,
          annotation=annotation)
      return data
    return None

  def parse_decorator(self, node):
    assert node.children[0].value == '@'
    name = self.name_to_string(node.children[1])
    call_expr = self.nodes_to_string(node.children[2:]).strip()
    return Decorator(name, Expression(call_expr) if call_expr else None)

  def parse_funcdef(self, parent, node, is_async, decorators):
    parameters = find(lambda x: x.type == syms.parameters, node.children)
    body = find(lambda x: x.type == syms.suite, node.children) or \
      find(lambda x: x.type == syms.simple_stmt, node.children)

    name = node.children[1].value
    docstring = self.get_docstring_from_first_node(body)
    args = self.parse_parameters(parameters)
    return_ = self.get_return_annotation(node)
    decorators = decorators or []

    return Function(self.location_from(node), parent, name, docstring,
      is_async=is_async, decorators=decorators, args=args, return_=return_)

  def parse_argument(self, node, argtype, scanner):
    # type: (Union[Leaf, Node], str, ListScanner) -> Argument
    """
    Parses an argument from the AST. *node* must be the current node at
    the current position of the *scanner*. The scanner is used to extract
    the optional default argument value that follows the *node*.
    """

    def parse_annotated_name(node):
      if node.type == syms.tname:
        scanner = ListScanner(node.children)
        name = scanner.current.value
        node = scanner.advance()
        assert node.type == token.COLON, node.parent
        node = scanner.advance()
        annotation = Expression(self.nodes_to_string([node]))
      elif node:
        name = node.value
        annotation = None
      else:
        raise RuntimeError('unexpected node: {!r}'.format(node))
      return (name, annotation)

    name, annotation = parse_annotated_name(node)

    node = scanner.advance()
    default = None
    if node and node.type == token.EQUAL:
      node = scanner.advance()
      default = Expression(self.nodes_to_string([node]))
      node = scanner.advance()

    return Argument(name, annotation, default, argtype)

  def parse_parameters(self, parameters):
    assert parameters.type == syms.parameters, parameters.type
    result = []

    arglist = find(lambda x: x.type == syms.typedargslist, parameters.children)
    if not arglist:
      # NOTE (@NiklasRosenstein): A single argument (annotated or not) does
      #   not get wrapped in a `typedargslist`, but in a single `tname`.
      tname = find(lambda x: x.type == syms.tname, parameters.children)
      if tname:
        scanner = ListScanner(parameters.children, parameters.children.index(tname))
        result.append(self.parse_argument(tname, Argument.POS, scanner))
      else:
        assert len(parameters.children) in (2, 3), parameters.children
        if len(parameters.children) == 3:
          result.append(Argument(parameters.children[1].value, None, None, Argument.POS))
      return result

    argtype = Argument.POS

    index = ListScanner(arglist.children)
    for node in index.safe_iter(auto_advance=False):
      node = index.current
      if node.type == token.STAR:
        node = index.advance()
        if node.type == token.COMMA:
          result.append(Argument('', None, None, Argument.KW_SEPARATOR))
          index.advance()
        else:
          result.append(self.parse_argument(node, Argument.POS_REMAINDER, index))
          index.advance()
        argtype = Argument.KW_ONLY
        continue
      elif node.type == token.DOUBLESTAR:
        node = index.advance()
        result.append(self.parse_argument(node, Argument.KW_REMAINDER, index))
        continue
      result.append(self.parse_argument(node, argtype, index))
      index.advance()

    return result

  def parse_classdef_arglist(self, classargs):
    metaclass = None
    bases = []
    for child in classargs.children[::2]:
      if child.type == syms.argument:
        key, value = child.children[0].value, self.nodes_to_string(child.children[2:])
        if key == 'metaclass':
          metaclass = Expression(value)
        else:
          # TODO @NiklasRosenstein: handle metaclass arguments
          pass
      else:
        bases.append(Expression(str(child)))
    return metaclass, bases

  def parse_classdef_rawargs(self, classdef):
    metaclass = None
    bases = []
    index = ListScanner(classdef.children, 2)
    if index.current.type == token.LPAR:
      index.advance()
      while index.current.type != token.RPAR:
        if index.current.type == syms.argument:
          key = index.current.children[0].value
          value = Expression(str(index.current.children[2]))
          if key == 'metaclass':
            metaclass = value
          else:
            # TODO @NiklasRosenstein: handle metaclass arguments
            pass
        else:
          bases.append(Expression(str(index.current)))
        index.advance()
    return metaclass, bases

  def parse_classdef(self, parent, node, decorators):
    name = node.children[1].value
    bases = []
    metaclass = None

    # An arglist is available if there are at least two parameters.
    # Otherwise we have to deal with parsing a raw sequence of nodes.
    classargs = find(lambda x: x.type == syms.arglist, node.children)
    if classargs:
      metaclass, bases = self.parse_classdef_arglist(classargs)
    else:
      metaclass, bases = self.parse_classdef_rawargs(node)

    suite = find(lambda x: x.type == syms.suite, node.children)
    docstring = self.get_docstring_from_first_node(suite)
    class_ = Class(self.location_from(node), parent, name, docstring,
      bases=bases, metaclass=metaclass, decorators=decorators)

    for child in suite.children:
      if isinstance(child, Node):
        member = self.parse_declaration(class_, child)
        if metaclass is None and isinstance(member, Data) and \
            member.name == '__metaclass__':
          metaclass = member.expr
          member.remove()

    return class_

  def location_from(self, node):
    return Location(self.filename, node.get_lineno())

  def get_return_annotation(self, node):
    rarrow = find(lambda x: x.type == token.RARROW, node.children)
    if rarrow:
      node = rarrow.next_sibling
      return Expression(self.nodes_to_string([node]))
    return None

  def get_most_recent_prefix(self, node):
    if node.prefix:
      return node.prefix
    while not node.prev_sibling and not node.prefix:
      if not node.parent:
        return ''
      node = node.parent
    if node.prefix:
      return node.prefix
    node = node.prev_sibling
    while isinstance(node, Node) and node.children:
      node = node.children[-1]
    return node.prefix

  def get_docstring_from_first_node(self, parent, module_level=False):
    """
    This method retrieves the docstring for the block node *parent*. The
    node either declares a class or function.
    """

    node = find(lambda x: isinstance(x, Node), parent.children)
    if node and node.type == syms.simple_stmt:
      if node.children[0].type == token.STRING:
        return self.prepare_docstring(node.children[0].value)
    if not node and not module_level:
      return None
    if self.treat_singleline_comment_blocks_as_docstrings:
      docstring, doc_type = self.get_hashtag_docstring_from_prefix(node or parent)
      if doc_type == 'block':
        return docstring
    return None

  def get_statement_docstring(self, node):
    prefix = self.get_most_recent_prefix(node)
    ws = re.match(r'\s*', prefix[::-1]).group(0)
    if ws.count('\n') == 1:
      docstring, doc_type = self.get_hashtag_docstring_from_prefix(node)
      if doc_type == 'statement':
        return docstring
    # Look for the next string literal instead.
    while node and node.type != syms.simple_stmt:
      node = node.parent
    if node and node.next_sibling and node.next_sibling.type == syms.simple_stmt:
      string_literal = node.next_sibling.children[0]
      if string_literal.type == token.STRING:
        return self.prepare_docstring(string_literal.value)
    return None

  def get_hashtag_docstring_from_prefix(self, node):
    # type: (Node) -> (Optional[str], Optional[str])
    """
    Given a node in the AST, this method retrieves the docstring from the
    closest prefix of this node (ie. any block of single-line comments that
    precede this node).

    The function will also return the type of docstring: A docstring that
    begins with `#:` is a statement docstring, otherwise it is a block
    docstring (and only used for classes/methods).

    return: (docstring, doc_type)
    """

    prefix = self.get_most_recent_prefix(node)
    lines = []
    doc_type = None
    for line in reversed(prefix.split('\n')):
      line = line.strip()
      if lines and not line.startswith('#'):
        break
      if doc_type is None and line.strip().startswith('#:'):
        doc_type = 'statement'
      elif doc_type is None and line.strip().startswith('#'):
        doc_type = 'block'
      if lines or line:
        lines.append(line)
    return self.prepare_docstring('\n'.join(reversed(lines))), doc_type

  def prepare_docstring(self, s):
    # TODO @NiklasRosenstein handle u/f prefixes of string literal?
    s = s.strip()
    if s.startswith('#'):
      lines = []
      for line in s.split('\n'):
        line = line.strip()
        if line.startswith('#:'):
          line = line[2:]
        else:
          line = line[1:]
        lines.append(line.lstrip())
      return '\n'.join(lines).strip()
    if s.startswith('"""') or s.startswith("'''"):
      return dedent_docstring(s[3:-3]).strip()
    if s.startswith('"') or s.startswith("'"):
      return dedent_docstring(s[1:-1]).strip()
    return None

  def nodes_to_string(self, nodes):
    """
    Converts a list of AST nodes to a string.
    """

    def generator(nodes, skip_prefix=True):
      # type: (List[Node], Bool) -> Iterable[Tuple[str, str]]
      for i, node in enumerate(nodes):
        if not skip_prefix or i != 0:
          yield node.prefix
        if isinstance(node, Node):
          for _ in generator(node.children, True):
            yield _
        else:
          yield node.value

    return ''.join(generator(nodes))

  def name_to_string(self, node):
    if node.type == syms.dotted_name:
      return ''.join(x.value for x in node.children)
    else:
      return node.value


class ListScanner:
  """
  A helper class to navigate through a list. This is useful if you would
  usually iterate over the list by index to be able to acces the next
  element during the iteration.

  Example:

  ```py
  scanner = ListScanner(lst)
  for value in scanner.safe_iter():
    if some_condition(value):
      value = scanner.advance()
  ```
  """

  def __init__(self, lst, index=0):
    self._list = lst
    self._index = index

  def __bool__(self):
    return self._index < len(self._list)

  __nonzero__ = __bool__

  @property
  def current(self):
    """
    Returns the current list element.
    """

    return self._list[self._index]

  def can_advance(self):
    """
    Returns `True` if there is a next element in the list.
    """

    return self._index < (len(self._list) - 1)

  def advance(self, expect=False):
    """
    Advances the scanner to the next element in the list. If *expect* is set
    to `True`, an #IndexError will be raised when there is no next element.
    Otherwise, `None` will be returned.
    """

    self._index += 1
    try:
      return self.current
    except IndexError:
      if expect:
        raise
      return None

  def safe_iter(self, auto_advance=True):
    """
    A useful generator function that iterates over every element in the list.
    You may call #advance() during the iteration to retrieve the next
    element in the list within a single iteration.

    If *auto_advance* is `True` (default), the function generator will
    always advance to the next element automatically. If it is set to `False`,
    #advance() must be called manually in every iteration to ensure that
    the scanner has advanced at least to the next element, or a
    #RuntimeError will be raised.
    """

    index = self._index
    while self:
      yield self.current
      if auto_advance:
        self.advance()
      elif self._index == index:
        raise RuntimeError('next() has not been called on the ListScanner')


@implements(Loader)
class PythonLoader(Struct):
  """
  This implementation of the #Loader interface parses Python modules and
  packages. Which files are parsed depends on the configuration (see
  #PythonLoaderConfig).

  If no #modules or #packages are set, the loader will attempt to find
  # packages in the src/ folder of the current directory.
  """

  #: A list of module names that this loader will search for and then parse.
  #: The modules are searched using the #sys.path of the current Python
  # interpreter, unless the #search_path option is specified.
  modules = Field([str], default=None)

  #: A list of package names that this loader will search for and then parse,
  #: including all sub-packages and modules.
  packages = Field([str], default=None)

  #: The module search path. If not specified, the current #sys.path is
  #: used instead. If any of the elements contain a `*` (star) symbol, it
  #: will be expanded with #sys.path.
  search_path = Field([str], default=None)

  #: The "print_function" flag will be passed down to the lib2to3
  #: RefactoringTool. This enables parsing Python 3 code that uses the
  #: print function without importing the print_function from the
  #: `__future__` module.
  print_function = Field(bool, default=True)

  #: Whether to treat blocks of single line comments as docstrings at the
  #: header of a module or inside a class or function definition. Defaults to True.
  treat_singleline_comment_blocks_as_docstrings = Field(bool, default=True)

  IGNORE_DISCOVERED_MODULES = ('test', 'tests', 'setup')

  def load(self, graph):
    if self.search_path is None:
      search_path = ['.', 'src'] if self.modules is None else sys.path
    else:
      search_path = list(self.search_path)
      if '*' in search_path:
        index = search_path.index('*')
        search_path[index:index+1] = sys.path

    if self.modules is None and self.packages is None:
      modules = []
      packages = []
      for path in search_path:
        try:
          items = os.listdir(path)
        except OSError:
          continue
        for name in items:
          if name.endswith('.py') and name[:-3] not in self.IGNORE_DISCOVERED_MODULES:
            modules.append(name[:-3])
            continue
          full_path = os.path.join(path, name, '__init__.py')
          if os.path.isfile(full_path) and name not in self.IGNORE_DISCOVERED_MODULES:
            packages.append(name)
      if not modules and not packages:
        logger.warning('No Python modules or packages detected.')
      else:
        logger.info('Detected packages in search_path: %s', packages)
    else:
      modules = self.modules or []
      packages = self.packages or []

    old_path = sys.path
    sys.path = search_path
    try:
      for module in modules:
        self._load_module(graph, module, False)
      for package in packages:
        self._load_module(graph, package, True)
    finally:
      sys.path = old_path

  def _load_module(self, graph: ModuleGraph, module_name: str, recursive: bool) -> None:
    try:
      loader = pkgutil.find_loader(module_name)
      if loader is None:
        raise LoaderError('module "{}" not found'.format(module_name))
      path = loader.get_filename()
    except ImportError as exc:
      raise LoaderError(exc)

    if recursive:
      if os.path.basename(path).startswith('__init__.'):
        path = os.path.dirname(path)
      for submodule_name, filename in self._iter_module_files(module_name, path):
        module = self.load_file(submodule_name, filename)
        graph.add_module(module)
    else:
      module = self.load_file(module_name, path)
      graph.add_module(module)

  def _iter_module_files(self, module_name, path):
    # pylint: disable=stop-iteration-return
    if os.path.isfile(path):
      yield module_name, path
      return
    elif os.path.isdir(path):
      yield next(self._iter_module_files(module_name, os.path.join(path, '__init__.py')))
      for item in os.listdir(path):
        if item == '__init__.py':
          continue
        item_abs = os.path.join(path, item)
        name = module_name + '.' + item.rstrip('.py')
        if os.path.isdir(item_abs) and os.path.isfile(os.path.join(item_abs, '__init__.py')):
          for x in self._iter_module_files(name, item_abs):
            yield x
        elif os.path.isfile(item_abs) and item_abs.endswith('.py'):
          yield next(self._iter_module_files(name, item_abs))
      return
    else:
      raise LoaderError('path "{}" does not exist'.format(path))

  def _get_parser(self):
    parser = Parser()
    parser.treat_singleline_comment_blocks_as_docstrings = self.treat_singleline_comment_blocks_as_docstrings
    return parser

  def load_file(self, module_name, filename):
    return parse_file(
      None,
      filename,
      module_name,
      parser=self._get_parser(),
      print_function=self.print_function)

  def load_source(self, source_code, module_name, filename=None):
    ast = parse_to_ast(
      source_code,
      filename,
      print_function=self.print_function)
    return self._get_parser().parse(ast, filename, module_name)



# This file was automatically generated by Shore. Do not edit manually.
# For more information on Shore see https://pypi.org/project/nr.shore/

from __future__ import print_function
import io
import os
import re
import setuptools
import sys

with io.open('src/pydoc_markdown/__init__.py', encoding='utf8') as fp:
  version = re.search(r"__version__\s*=\s*'(.*)'", fp.read()).group(1)

readme_file = 'README.md'
if os.path.isfile(readme_file):
  with io.open(readme_file, encoding='utf8') as fp:
    long_description = fp.read()
else:
  print("warning: file \"{}\" does not exist.".format(readme_file), file=sys.stderr)
  long_description = None

requirements = ['click >=7.0,<8.0.0', 'nr.collections >=0.0.1,<0.1.0', 'nr.databind.core >=0.0.18,<0.1.0', 'nr.databind.json >=0.0.9,<0.1.0', 'six >=1.11.0,<2.0.0', 'PyYAML >=5.3,<6.0.0', 'watchdog >=0.10.2,<1.0.0']

setuptools.setup(
  name = 'pydoc-markdown',
  version = version,
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'Create Python API documentation in Markdown format.',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/NiklasRosenstein/pydoc-markdown',
  license = 'MIT',
  packages = setuptools.find_packages('src', ['test', 'test.*', 'docs', 'docs.*']),
  package_dir = {'': 'src'},
  include_package_data = True,
  install_requires = requirements,
  extras_require = {},
  tests_require = [],
  python_requires = None, # TODO: '>=3.6,<4.0.0',
  data_files = [],
  entry_points = {
    'console_scripts': [
      'pydoc-markdown = pydoc_markdown.main:cli',
    ],
    'pydoc_markdown.interfaces.Loader': [
      'python = pydoc_markdown.contrib.loaders.python:PythonLoader',
    ],
    'pydoc_markdown.interfaces.Processor': [
      'crossref = pydoc_markdown.contrib.processors.crossref:CrossrefProcessor',
      'filter = pydoc_markdown.contrib.processors.filter:FilterProcessor',
      'google = pydoc_markdown.contrib.processors.google:GoogleProcessor',
      'pydocmd = pydoc_markdown.contrib.processors.pydocmd:PydocmdProcessor',
      'smart = pydoc_markdown.contrib.processors.smart:SmartProcessor',
      'sphinx = pydoc_markdown.contrib.processors.sphinx:SphinxProcessor',
    ],
    'pydoc_markdown.interfaces.Renderer': [
      'markdown = pydoc_markdown.contrib.renderers.markdown:MarkdownRenderer',
      'mkdocs = pydoc_markdown.contrib.renderers.mkdocs:MkdocsRenderer',
    ]
  },
  cmdclass = {},
  keywords = ['documentation', 'docs', 'generator', 'markdown', 'pydoc'],
  classifiers = ['Development Status :: 3 - Alpha', 'Intended Audience :: Developers', 'Intended Audience :: End Users/Desktop', 'Topic :: Software Development :: Code Generators', 'Topic :: Utilities', 'License :: OSI Approved :: MIT License', 'Programming Language :: Python :: 2', 'Programming Language :: Python :: 2.7', 'Programming Language :: Python :: 3', 'Programming Language :: Python :: 3.3', 'Programming Language :: Python :: 3.4', 'Programming Language :: Python :: 3.5'],
)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" Implements the `mkdocs` renderer. It uses the `markdown` renderer to
produce an MkDocs-compatible folder structure. """

from nr.databind.core import Field, Struct, ProxyType
from nr.interface import implements, override
from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer
from pydoc_markdown.interfaces import Renderer, Resolver, Server
from pydoc_markdown.reflection import Object, ModuleGraph
from pydoc_markdown.util.pages import Page, Pages
from typing import Dict, Iterable, List, Optional, Tuple
import copy
import fnmatch
import logging
import os
import re
import shutil
import subprocess
import yaml

logger = logging.getLogger(__name__)


class CustomizedMarkdownRenderer(MarkdownRenderer):
  """ We override some defaults in this subclass. """

  render_toc = Field(bool, default=False)


@implements(Renderer, Server)
class MkdocsRenderer(Struct):
  #: The output directory for the generated Markdown files. Defaults to
  #: `build/docs`.
  output_directory = Field(str, default='build/docs')

  #: Name of the content directory. Defaults to "content".
  content_directory_name = Field(str, default='content')

  #: Delete the content directory in the #output_directory before rendering.
  #: Note that this does not play well with pre-existing content, as it will
  #: be deleted as well. Defaults to False.
  clean_render = Field(bool, default=False)

  #: The pages to render into the output directory.
  pages = Field(Pages)

  #: Markdown renderer settings.
  markdown = Field(CustomizedMarkdownRenderer, default=CustomizedMarkdownRenderer)

  #: The name of the site. This will be carried into the `site_name` key
  #: of the #mkdocs_config.
  site_name = Field(str, default=None)

  #: Arbitrary configuration values that will be rendered to an
  #: `mkdocs.yml` file.
  mkdocs_config = Field(dict, default=dict, nullable=True)

  @property
  def content_dir(self) -> str:
    return os.path.join(self.output_directory, self.content_directory_name)

  def generate_mkdocs_nav(self, page_to_filename: Dict[Page, str]) -> Dict:
    def _generate(pages):
      result = []
      for page in pages:
        if page.children:
          result.append({page.title: _generate(page.children)})
        elif page.href:
          result.append({page.title: page.href})
        else:
          filename = os.path.relpath(page_to_filename[id(page)], self.content_dir)
          result.append({page.title: filename})
      return result
    return _generate(self.pages)

  # Renderer

  @override
  def render(self, graph: ModuleGraph) -> None:
    if self.clean_render and os.path.isdir(self.content_dir):
      logger.info('Cleaning directory "%s"', self.content_dir)
      shutil.rmtree(self.content_dir)

    page_to_filename = {}

    for item in self.pages.iter_hierarchy():
      filename = item.filename(self.content_dir, '.md')
      if not filename:
        continue

      page_to_filename[id(item.page)] = filename
      self.markdown.filename = filename
      item.page.render(filename, graph, self.markdown)

    config = copy.deepcopy(self.mkdocs_config)
    if self.site_name:
      config['site_name'] = self.site_name
    if not config.get('site_name'):
      config['site_name'] = 'My Project'
    config['docs_dir'] = self.content_directory_name
    config['nav'] = self.generate_mkdocs_nav(page_to_filename)

    if self.mkdocs_config is not None:
      filename = os.path.join(self.output_directory, 'mkdocs.yml')
      logger.info('Rendering "%s"', filename)
      with open(filename, 'w') as fp:
        yaml.dump(config, fp)

  @override
  def get_resolver(self, graph: ModuleGraph) -> Optional[Resolver]:
    # TODO (@NiklasRosenstein): The resolver returned by the Markdown
    #   renderer does not implement linking across multiple pages.
    return self.markdown.get_resolver(graph)

  # Server

  @override
  def get_server_url(self) -> str:
    return 'http://localhost:8000'

  @override
  def start_server(self) -> subprocess.Popen:
    return subprocess.Popen(['mkdocs', 'serve'], cwd=self.output_directory)

  @override
  def reload_server(self, process: subprocess.Popen):
    # While MkDocs does support file watching and automatic reloading,
    # it appears to be a bit quirky. Let's just restart the whole process.
    process.terminate()



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Field, Struct
from nr.interface import implements
from pydoc_markdown.interfaces import Processor
from pydoc_markdown.contrib.processors.google import GoogleProcessor
from pydoc_markdown.contrib.processors.pydocmd import PydocmdProcessor
from pydoc_markdown.contrib.processors.sphinx import SphinxProcessor


@implements(Processor)
class SmartProcessor(Struct):
  google = Field(GoogleProcessor, default=GoogleProcessor)
  pydocmd = Field(PydocmdProcessor, default=PydocmdProcessor)
  sphinx = Field(SphinxProcessor, default=SphinxProcessor)

  def process(self, graph, _resolver):
    """
    Preprocesses a given section into its components.
    """

    graph.visit(self.process_node)

  def process_node(self, node):
    if not node.docstring:
      return None
    if self.sphinx.check_docstring_format(node.docstring):
      return self.sphinx.process_node(node)
    if self.google.check_docstring_format(node.docstring):
      return self.google.process_node(node)
    return self.pydocmd.process_node(node)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" Provides the #CrossrefProcessor that replaces cross references in
docstrings with actual hyperlinks. """

from nr.databind.core import Struct
from nr.interface import implements, override
from pydoc_markdown.interfaces import Processor
import logging
import re

logger = logging.getLogger(__name__)


@implements(Processor)
class CrossrefProcessor(Struct):

  @override
  def process(self, graph, resolver):
    graph.visit(lambda x: self._preprocess_refs(x, resolver))

  def _preprocess_refs(self, node, resolver):
    if not resolver or not node.docstring:
      return

    def handler(match):
      ref = match.group('ref')
      parens = match.group('parens') or ''
      trailing = (match.group('trailing') or '').lstrip('#')
      # Remove the dot from the ref if its trailing (it is probably just
      # the end of the sentence).
      has_trailing_dot = False
      if trailing and trailing.endswith('.'):
        trailing = trailing[:-1]
        has_trailing_dot = True
      elif not parens and ref.endswith('.'):
        ref = ref[:-1]
        has_trailing_dot = True
      href = resolver.resolve_ref(node, ref)
      if href:
        result = '[`{}`]({})'.format(ref + parens + trailing, href)
      else:
        logger.warning('ref "%s" could not be resolved in %s "%s" (%s)',
          ref, type(node).__name__, node.path(), node.location)
        result = '`{}`'.format(ref + parens)
      # Add back the dot.
      if has_trailing_dot:
        result += '.'
      return result

    node.docstring = re.sub(
      r'\B#(?P<ref>[\w\d\._]+)(?P<parens>\(\))?(?P<trailing>#[\w\d\._]+)?',
      handler,
      node.docstring)
